# Socket server in python using select function
 
import socket, select, netifaces, sys, getopt, os
import signal
import time
import re
from ctypes import cdll

# Constant taken from http://linux.die.net/include/linux/prctl.h
PR_SET_PDEATHSIG = 1

class PrCtlError(Exception):
   pass

def on_parent_exit():
   signum = 9

   # http://linux.die.net/man/2/prctl
   result = cdll['libc.so.6'].prctl(PR_SET_PDEATHSIG, signum)
   if result != 0:
      raise PrCtlError('prctl failed with error code %s' % result)

GRE_DEV_NAME = "tcr_gre"
PROTO = 155
PROBE_INTERVAL = 1.0
TIMEOUT = 5.0

gre_dic = {}

def gre_dev_alloc_name():
   ifs = netifaces.interfaces()
   mark = 1
   while GRE_DEV_NAME + str(mark - 1) in ifs:
      mark += 1

   return GRE_DEV_NAME + str(mark - 1), mark

def ip2tunnel(ip):
   return os.popen("ip tunnel show remote " + ip + " | awk -F: '{ print $1 }' | head -1").read().strip()

def fill_gre_devs():
   now = time.time()
   ifs = netifaces.interfaces()
   for i in ifs:
      if GRE_DEV_NAME in i:
         gre_dic[i] = now

def clear_gre_devs():
   print "Clearing GRE"
   clear_list = []
   now = time.time()
   for key, value in gre_dic.iteritems():
      print "now: " + str(now) + " value: " + str(value) + " TIMEOUT: " + str(TIMEOUT) + " " + str(now - value > TIMEOUT)
      if now - value > TIMEOUT:
         clear_list.append(key)

   for dev in clear_list:
      del gre_dic[dev]
      print "Removing " + dev
      os.system("ip tunnel del " + dev)
      mark = int(re.sub(GRE_DEV_NAME, '', dev)) + 1
      os.system("iptables -t mangle -D FORWARD -m state --state NEW -i " + dev + " -j CONNMARK --set-mark " + str(mark))
      os.system("ip rule del fwmark " + str(mark) + " pref " + str(PROTO))
   
def hit_gre_dev(ip):
   dev = ip2tunnel(ip)
   print "Hitting " + dev
   gre_dic[dev] = time.time()

if __name__ == "__main__":
   port = 5000
   timeout = 5.0

   on_parent_exit()

   try:
      opts, args = getopt.getopt(sys.argv[1:],"hp:",["port="])
   except getopt.GetoptError as err:
      print str(err)
      usage()
      sys.exit(2)

   for opt, arg in opts:
      if opt == '-h':
         usage()
         sys.exit()
      elif opt in ("-p", "--port"):
         port = int(arg)

   if port <= 0 or port > 65535:
      print "Missing or bad port value"
      usage()
      sys.exit(2)

   fill_gre_devs()

   server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
   server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
   server_socket.bind(("0.0.0.0", port))
   server_socket.listen(10)
   server_socket.setblocking(1)

   print "Dynamic GRE server started on port " + str(port)

   select_read = [server_socket]
   select_error = []

   t0 = time.time()
   while 1:
      t1 = time.time()
      delta = PROBE_INTERVAL - (t1 - t0)
      if delta < 0.0:
         delta = 0.0
      # Get the list sockets which are ready to be read through select
      read_sockets,write_sockets,error_sockets = select.select(select_read,[],select_error,delta)
      
      if not read_sockets and not write_sockets and not error_sockets:
         t0 = t1
         clear_gre_devs()
         continue

      selected = read_sockets + error_sockets
      for sock in selected:
         #New connection
         if sock == server_socket:
            # Handle the case in which there is a new connection recieved through server_socket
            sockfd, remote = server_socket.accept()
            local = sockfd.getsockname()
            print "Client (%s, %s) connected" % remote
             
            new_dev, mark = gre_dev_alloc_name()
            if os.system("test $(ip tunnel show remote " + remote[0] + " local " + local[0] + " | wc -l) -eq 0") == 0:
               os.system("ip tunnel add " + new_dev + " mode gre remote " + remote[0] + " local " + local[0] + " ttl 255")
               os.system("ifconfig " + new_dev + " up")
               os.system("iptables -t mangle -A FORWARD -m state --state NEW -i " + new_dev + " -j CONNMARK --set-mark " + str(mark))
               os.system("ip rule add fwmark " + str(mark) + " table " + str(mark) + " pref " + str(PROTO))
               os.system("ip route add default table " + str(mark) + " dev " + new_dev + " proto " + str(PROTO))
            select_read.append(sockfd)
            select_error.append(sockfd)
            hit_gre_dev(remote[0])
         else:
               print "Closing (%s, %s)" % sock.getpeername()
               sock.close()
               select_read.remove(sock)
               select_error.remove(sock)
   server_socket.close()
